/*
 
 Copyright 1996 by J.D. Koftinoff Software Ltd, 
 under the KPL (Koftinoff Public License)
 
 THE NO-NONSENSE KOFTINOFF PUBLIC LICENSE (KPL)

 
 This software is provided free of charge.  You use this software only under 
 the following conditions:
 
    * You agree that you use this software at your own risk
    * You agree that the copyright of this software is held 
      by J.D. Koftinoff Software, Ltd.
 
 If you agree to these conditions, you may modify and use the software in 
 commercial as well as free projects.  However, If you use this software in a 
 commercial project, you MUST send a neat postcard to me at:
 
    Jeff Koftinoff
    J.D. Koftinoff Software, Ltd.
    #5 - 1131 Burnaby St.
    Vancouver, B.C.
    Canada
    V6E-1P3
 
    email: jeffk@jdkoftinoff.com
 
 Have Fun!
 
 P.S. Send me a postcard anyways please! I'll send one back!
 */

#include "jkcthread.h"
#include "jkthread.h"


//--------------------------------------------------------------------------------
JKRecursiveSemaphore sema;
JKSemaphore semb(0);

class Test1 : public JKThread
{
public:
	Test1(int n_)
	{
	    n=n_;
	}

	// All thread subclass destructors MUST call ThreadKill()
	// This ensures the thread gets killed if the thread object is
	// getting deleted by a different thread
	
	virtual ~Test1()
	{
	    ThreadKill();	
	}
	
	
protected:  
	void PreMain();
	void Main();
    void PostMain();
	
	int n;
};


void Test1::PreMain()
{
	printf( "\n%d: About to Run Main\n", ThreadGetPID() );
}

void Test1::PostMain()
{
	printf( "\n%d: Finished Main\n", ThreadGetPID() );
}


void Test1::Main()
{
	char b[80];
	int i;
	int id=ThreadGetPID();
	
	for( i=0; i<n+2; ++i )
	{
	    b[i]='\t';
	}
	b[i]='\0';
	
	printf( "%d: %s get semb\n", id, b );
	
	semb.Get();
	
	printf( "%d: %s Got semb\n", id, b );
	
    for( i=0; i<4; ++i )
	{
		char *p;
		
	    sleep( 1 );

		/* try mallocing */
		
		p=(char *)malloc(1000);
		
		printf( "%d: %s trying\n", id, b );
		if( sema.Get() )
		{
			/* we got it! now sleep and release it */
		    printf( "%d: %s got\n",id,b );
			
			sleep(n+1);
			printf( "%d: %s release\n",id,b );
			
			sema.Release();

			/* free the memory we malloced */
			
			free(p);
		}
		else
		{
		    printf( "%d: %s Can't get\n",id,b );	
		}				
	}
    
}


//--------------------------------------------------------------------------------


class Test2 : public JKThreadWithMsgQ
{
public:

	Test2()
	{
	    dest=0;	
	}

	// All thread subclass destructors MUST call ThreadKill()
	// This ensures the thread gets killed if the thread object is
	// getting deleted by a different thread
	
	virtual ~Test2()
	{
	    ThreadKill();	
	}
	
	void SetDest( JKThreadWithMsgQ *dest_ )
	{
	    dest=dest_;	
	}
	
	
protected:  
	void Main();

	JKThreadWithMsgQ *dest;
};


void Test2::Main()
{
	JKMsg m;
	
    while(1)
	{		
    	if( ThreadGetMsg( &m ) )
		{		  
            printf( "%5d: Got message: %ld, %ld, %ld\n", ThreadGetPID(), m.type, m.param1, m.param2 );

    		if( dest )
    		{			
                printf( "%5d: Sending to : %ld\n", ThreadGetPID(), dest->ThreadGetPID() );		
        		dest->ThreadPostMsg( m.type, m.param1, m.param2+1 );
    		}
			else
			{
			    printf( "%5d: no destination\n\n", ThreadGetPID() );	
			}
			
		}
		else
		{
			printf( "%5d: Error getting message!\n", ThreadGetPID() );
		}
		
	}
	
}


int main()
{

	printf( "\nMAIN: Doing test 1\n" );
	
	{		
        Test1 a(0),b(1),c(2),d(3);	
	
    	a.ThreadRun();
    	b.ThreadRun();
    	c.ThreadRun();
    	d.ThreadRun();
		
		/* semb was inited with a count value of 0 - 
		 * This means any threads waiting for it will block
		 * until we release it once for each thread 
		 */
		
		sleep(1);
		printf( "MAIN: Releasing semb\n" );
		semb.Release();
		sleep(4);
		printf( "MAIN: Releasing semb\n" );		
		semb.Release();
	    sleep(4);
		printf( "MAIN: Releasing semb\n" );		
		semb.Release();
		sleep(4);
		printf( "MAIN: Releasing semb\n" );		
		semb.Release();
		
    	sleep(25);
		
		// kill one of the threads
		
		printf( "\nMAIN: KILLING D %d\n", d.ThreadGetPID() );
		d.ThreadKill();
		
		sleep(10);
		
		// we really should kill the threads before they go out of scope.
        // the destructor will do this, but only after the objects
		// are partially destroyed.

		printf( "\nMAIN: Killing all Test1 threads\n" );
		
		a.ThreadKill();
		b.ThreadKill();
		c.ThreadKill();
		d.ThreadKill();
	}	

	printf( "\nMAIN: Doing test 2\n" );
	
	{
	    Test2 a,b,c,d;
		
		a.SetDest( &b );
		b.SetDest( &c );
		c.SetDest( &d );
		d.SetDest( 0 );
		
		a.ThreadRun();
		b.ThreadRun();
		c.ThreadRun();
		d.ThreadRun();		
		
		
		for( int i=0; i<12; ++i )
		{	
            sleep(1);

			// send a message to the first thread. It will modify it 
		    // and copy it up the line.
			
			printf( "\nMAIN: Sending %d, %d, %d\n", 1,i,0 );
            a.ThreadPostMsg( 1, i, 0 );					
		}

		printf( "\nMAIN: Killing Test2 Threads\n" );
		a.ThreadKill();
		b.ThreadKill();
		c.ThreadKill();
		d.ThreadKill();
	}

	printf( "\nMAIN: EXITING\n" );
}


